/******************************************************************************

	add.c

	Simulate HP9845 BPC binary integer add instructions (ADA, ADB).
	
	Usage: add <arg1> <arg2>

	(c) 2006 by Ansgar Kueckes (kueckes@alice-dsl.net)

******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

void bpc_unsigned_add(unsigned long *reg, unsigned long summand, int *extend, int *overflow)
{
  long tmp,a,b;
  unsigned long result;

  /* check for overflow condition */  
  a=*reg; if (a>32767) a-=65536;
  b=summand; if (b>32767) b-=65536;
  tmp=a+b;
  if ((tmp<-32768) || (tmp>32767)) *overflow=1;

  /* integer add */
  result=(*reg)+summand;
  
  /* carry from bit 15 ? */
  if (result>65535) *extend=1;
  	
  *reg=result&0xFFFF;
}

void bpc_signed_add(long *reg, long summand, int *extend, int *overflow)
{
  long tmp,a,b,result;

  /* check for overflow condition */  
  tmp=(*reg)+summand;
  if ((tmp<-32768) || (tmp>32767)) *overflow=1;

  /* integer add */
  a=*reg; if (a<0) a+=65536; 
  b=summand; if (b<0) b+=65536;
  result=a+b;
  
  /* carry from bit 15 ? */
  if (result>65535) { result-=65536; *extend=1; }

  if (result>32767) result-=65536;
  *reg=result;
}

int main(int argc, char *argv[])
{
  int overflow, extend;
  long arg1,arg2;

  arg1=atoi(argv[1]);
  arg2=atoi(argv[2]);
  overflow=extend=0;
  bpc_unsigned_add(&arg1, arg2, &extend, &overflow);
  printf("Unsigned result(1): %ld Ex: %d Ov: %d\n", arg1, extend, overflow);
  while(extend) { extend=0; bpc_unsigned_add(&arg1, 1, &extend, &overflow); }
  printf("         result(2): %ld Ex: %d Ov: %d\n\n", arg1, extend, overflow);

  arg1=atoi(argv[1]);
  arg2=atoi(argv[2]);
  overflow=extend=0;
  bpc_signed_add(&arg1, arg2, &extend, &overflow);
  printf("  Signed result(1): %ld Ex: %d Ov: %d\n", arg1, extend, overflow);
  while(extend) { extend=0; bpc_signed_add(&arg1, 1, &extend, &overflow); }
  printf("         result(2): %ld Ex: %d Ov: %d\n", arg1, extend, overflow);
  
  exit(0);
}
