/******************************************************************************

	hp2bas.c

	Strips the pure BASIC code from HP 9845 Basic files (string type, saved
	with "SAVE")
	
	Note: beta version, not thoroughly tested, no deep error processing.

	(c) 2006 by Ansgar Kueckes (kueckes@alice-dsl.net)

******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>

void print_usage(char *command)
{
	printf("Usage: %s <input file> [<output file>]\n", command);
	printf("        <input file>  HP 9845 basic file\n");
	printf("        -v            verbose mode\n");
	return;
}


int main(int argc, char *argv[])
{
  FILE *ifp, *ofp = 0;
  size_t bytes_read;
  char *outfile = 0;
  char *infile = 0;
  unsigned char buffer[1024];
  unsigned char c;
  int rec_size = 256;
  int i, count, l;
  int verbose = 0;

  if (argc < 2) { print_usage(argv[0]); exit(0); }

  for ( count = 1; count < argc ; count++)
  {
	if (!strcmp(argv[count], "-v") && (count < argc))
		{ verbose = 1; continue; }

	if (!strncmp(argv[count], "-", 1) && (count < argc))
		 { print_usage(argv[0]); exit(0); }

	if (!infile) infile = argv[count];
	else outfile = argv[count];
  }
  if (!infile) { print_usage(argv[0]); exit(0); }
  
  ifp = fopen( infile, "rb");
  if (!ifp) { fprintf(stderr, "Could not open file \"%s\".\n", infile); exit(-1); }

  if (outfile) {
     ofp = fopen( outfile, "wb");
     if (!ofp) { fprintf(stderr, "Could not open file \"%s\" for writing.\n", outfile); exit(-1); }
  }
  else ofp = stdout;

  while ((bytes_read = fread((void *)buffer, 1, rec_size, ifp)) == rec_size) {
        if (ferror(ifp)) {
   	      fprintf(stderr, "Read error on binary file (%d bytes read).\n", bytes_read);
   	      exit(-1);
   	}
   	i = 0;
  	while (i < (bytes_read - 1)) {
  	    if (buffer[i] == 0x0d) {
  	    	i++;
  	    }
	    else if (buffer[i] == 0x00 && (buffer[i+1] == 0x3c || buffer[i+1] == 0x2c)) {
	    	/* get complete string */
	    	i += 2;
			l = buffer[i] * 256 + buffer[i+1];
			i += 2;
			for (count = 0; count < l; count++) {
				c = buffer[i++] & 0x7f;
				if (c < 0x20) c = 0x20;
				fputc(c, ofp);
			}
			fputc(0x0d, ofp);
			fputc(0x0a, ofp);
	    }
	    else if (buffer[i] == 0x00 && (buffer[i+1] == 0x1c || buffer[i+1] == 0x0c)) {
	    	/* get complete string */
	    	i = i + 2;
			l = buffer[i] * 256 + buffer[i+1];
			i = i + 2;
			while (i < rec_size) {
				c = buffer[i++] & 0x7f;
				if (c < 0x20) c = 0x20;
				fputc(c, ofp);
			}
	    } 
	    else if (buffer[i] == 0x00 && buffer[i+1] == 0x1e) break; 	/* EOR */
  	    else if (buffer[i] == 0x00 && buffer[i+1] == 0x3e) exit(1); /* EOF */
  	}
  }

   if (ifp) fclose(ifp);
   if (ofp) fclose(ofp);
   exit(0);
}
