/******************************************************************************

	progalyze.c

	PROG & BINARY file analyzer
	
	(c) 2008 by Ansgar Kueckes (ansgar@hp9845.net)

******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <windows.h>

#define READ_WORD(ret_val)	{ \
							if (fread(buf, 1, 2, ifp) != 2) { \
								fprintf(stderr, "Invalid input file (EOF detected during read)\n"); \
								exit(-1); \
						  	} \
							else { \
								ret_val = buf[0]*256 + buf[1]; \
								offset++; \
							} \
							}

#define READ_WRITE_WORD(ret_val)	{ \
							if (fread(buf, 1, 2, ifp) != 2) { \
								fprintf(stderr, "Invalid input file (EOF detected during read)\n"); \
								exit(-1); \
						  	} \
							else { \
								if (ofp) fwrite(buf, 1, 2, ofp); \
								ret_val = buf[0]*256 + buf[1]; \
								offset++; \
							} \
							}

#define WRITE_WORD(val)		if (ofp) { \
								buf[0] = val / 256; \
								buf[1] = val % 256; \
								fwrite(buf, 1, 2, ofp); \
							}
	
#define LOG 				print_log

/* globals */
int loglevel = 1;
int offset = 0;
BYTE buf[2];
FILE *ifp = 0;
FILE *ofp = 0;
WORD initial_offset = 0;
BOOL extract = FALSE;
BOOL unsecure = FALSE;
BOOL is_rom = FALSE;
char forbidden_characters[] = "*.\"/\\[]:;|=,";

static struct {
	WORD address;
	char *cmd;
} cmd_table[] = {

	// --- STANDARD ---
	{ 0x4B00, "BEEP" },
	{ 0x612B, "CALL" },
	{ 0x24C4, "COM" },
	{ 0x5398, "DATA" },
	{ 0x7EA6, "DEFAULT OFF" },
	{ 0x7E99, "DEFAULT ON" },
	{ 0x613A, "DEF FN" },
	{ 0x7C7D, "DEG" },
	{ 0x253D, "DIM" },
	{ 0x7E1C, "DISABLE" },
	{ 0x4A8F, "DISP" },
	{ 0x4FC0, "EDIT" },
	{ 0x4B33, "EDIT LINE" },
	{ 0x7DF3, "ENABLE" },
	{ 0x7C74, "END" },
	{ 0x7CA3, "FIXED" },
	{ 0x7CBE, "FLOAT" },
	{ 0x60E3, "FNEND" },
	{ 0x5FA2, "FOR" },
	{ 0x7EFA, "GOSUB" },
	{ 0x7EF5, "GO SUB" },
	{ 0x7EE6, "GOTO" },
	{ 0x7EE1, "GO TO" },
	{ 0x7C8F, "GRAD" },
	{ 0x7BF9, "IF" },
	{ 0x481E, "IMAGE" },
	{ 0x4FDF, "INPUT" },
	{ 0x258F, "INTEGER" },
	{ 0x7F74, "LET" },
	{ 0x7FDE, "=" },
	{ 0x4FD0, "LINPUT" },
	{ 0x6055, "NEXT" },
	{ 0x59F3, "NORMAL" },
	{ 0x7E8C, "OFF ERROR" },
	{ 0x7DDF, "OFF KEY #" },
	{ 0x7E28, "ON ERROR" },
	{ 0x7F21, "ON" },
	{ 0x7DB2, "ON KEY #" },
	{ 0x239C, "OPTION BASE" },
	{ 0x4B24, "OVERLAP" },
	{ 0x2102, "PAUSE" },
	{ 0x49DE, "PRINT" },
	{ 0x49CA, "PRINT ALL IS" },
	{ 0x4730, "PRINT USING" },
	{ 0x4967, "PRINTER IS" },
	{ 0x7C86, "RAD" },
	{ 0x7EB2, "RANDOMIZE" },
	{ 0x52AC, "READ" },
	{ 0x257B, "REAL" },
	{ 0x7B2B, "REDIM" },
	{ 0x7CCA, "REM" },
	{ 0x5372, "RESTORE" },
	{ 0x60D9, "RESUME INTERACTIVE" },
	{ 0x7C3F, "RETURN" },
	{ 0x5B7C, "SECURE" },
	{ 0x4B11, "SERIAL" },
	{ 0x2584, "SHORT" },
	{ 0x7C9A, "STANDARD" },
	{ 0x210D, "STOP" },
	{ 0x617B, "SUB" },
	{ 0x60EC, "SUBEND" },
	{ 0x60F5, "SUBEXIT" },
	{ 0x60C9, "SUSPEND INTERACTIVE" },
	{ 0x5B51, "TRACE" },
	{ 0x5AA9, "TRACE ALL" },
	{ 0x5A76, "TRACE ALL VARIABLES" },
	{ 0x5A34, "TRACE PAUSE" },
	{ 0x5ABC, "TRACE VARIABLES" },
	{ 0x5A24, "TRACE WAIT" },
	{ 0x611A, "TYPEWRITER OFF" },
	{ 0x610B, "TYPEWRITER ON" },
	{ 0x4AE9, "WAIT" },
	{ 0x7CFC, "!" },
	
	// --- MATRIX OPERATIONS ---
	{ 0x6FEA, "MAT" },
	{ 0x6FBC, "MAT PRINT" },
	{ 0x6FD3, "MAT READ" },
	{ 0x6FEA, "MAT =" },
	
	// --- MASS STORAGE ---
	{ 0x11BF, "ASSIGN" },
	{ 0x1147, "ASSIGN #" },
	{ 0x1544, "BUFFER #" },
	{ 0x1479, "CAT" },
	{ 0x130C, "CHECK READ" },
	{ 0x12F5, "CHECK READ #" },
	{ 0x12E8, "CHECK READ OFF" },
	{ 0x12C2, "CHECK READ OFF #" },
	{ 0x136F, "COPY" },
	{ 0x0E81, "CREATE" },
	{ 0x16BA, "GET" },
	{ 0x1418, "INITIALIZE" },
	{ 0x16CF, "LINK" },
	{ 0x176A, "LOAD ALL" },
	{ 0x178E, "LOAD BIN" },
	{ 0x17BF, "LOAD KEY" },
	{ 0x17E4, "LOAD" },
	{ 0x13CD, "MASS STORAGE IS" },
	{ 0x6FB0, "MAT PRINT #" },
	{ 0x6FC8, "MAT READ #" },
	{ 0x7DEA, "OFF END #" },
	{ 0x7DBC, "ON END #" },
	{ 0x4630, "PRINT #" },
	{ 0x0F26, "PROTECT" },
	{ 0x0EFE, "PURGE" },
	{ 0x46D4, "READ #" },
	{ 0x1325, "RENAME" },
	{ 0x1743, "RE-SAVE" },
	{ 0x182C, "RE-STORE" },
	{ 0x1451, "REWIND" },
	{ 0x171A, "SAVE" },
	{ 0x1819, "STORE" },
	{ 0x177A, "STORE ALL" },
	{ 0x17A4, "STORE BIN" },
	{ 0x17D3, "STORE KEY" },
	{ 0, 0 }
};


static struct {
	WORD block;
	WORD address;
	char *cmd;
} option_table[] = {
	
	// --- SYSTEM PPU ---
	{ 0x0005, 0x23B8, "OFF KBD" },
	{ 0x0005, 0x231E, "ON KBD CALL" },
	{ 0x0005, 0x230E, "ON KBD GOSUB" },
	{ 0x0005, 0x22FD, "ON KBD GOTO" },
	
	// --- MASS STORAGE ROM ---
	{ 0x0005, 0x2188, "CAT TO" },
	{ 0x0005, 0x2019, "FCREATE" },
	{ 0x0005, 0x20D0, "FPRINT" },
	{ 0x0005, 0x206E, "FREAD" },
	
	// --- GRAPHICS ROM ---
    { 0x0007, 0x330E, "ALPHA" },
    { 0x0007, 0x3388, "AREA COLOR" },
    { 0x0007, 0x3395, "AREA INTENSITY" },
	{ 0x0007, 0x3064, "AXES" },
	{ 0x0007, 0x3054, "CLIP" },
    { 0x0007, 0x33B9, "CONVERGE" },
	{ 0x0007, 0x3119, "CSIZE" },
	{ 0x0007, 0x3136, "CURSOR" },
    { 0x0007, 0x33C2, "DEGAUSS" },
	{ 0x0007, 0x312D, "DIGITIZE" },
	{ 0x0007, 0x30A4, "DRAW" },
	{ 0x0007, 0x315F, "DUMP GRAPHICS" },
	{ 0x0007, 0x32F0, "DUMP GRAPHICS #" },
    { 0x0007, 0x3318, "EXIT ALPHA" },
	{ 0x0007, 0x3153, "EXIT GRAPHICS" },
    { 0x0007, 0x320C, "FILL" },
	{ 0x0007, 0x310A, "FRAME" },
	{ 0x0007, 0x30F2, "GCLEAR" },
	{ 0x0007, 0x3167, "GLOAD" },
	{ 0x0007, 0x3147, "GRAPHICS" },
	{ 0x0007, 0x6618, "GRAPHICS INPUT" },
	{ 0x0007, 0x660B, "GRAPHICS INPUT IS" },
	{ 0x0007, 0x306C, "GRID" },
    { 0x0007, 0x3209, "GSTAT" },
	{ 0x0007, 0x3170, "GSTORE" },
	{ 0x0007, 0x307C, "IPLOT" },
    { 0x0007, 0x33B0, "KEY LABELS" },
	{ 0x0007, 0x30D1, "LABEL" },
    { 0x0007, 0x32FB, "LABEL KEY #" },
    { 0x0007, 0x3306, "LABEL KEYS" },
    { 0x0007, 0x30C1, "LABEL USING" },
    { 0x0007, 0x339D, "LAXES" },
	{ 0x0007, 0x3094, "LDIR" },
	{ 0x0007, 0x3094, "LDIR" },
	{ 0x0007, 0x30E1, "LETTER" },
	{ 0x0007, 0x305C, "LIMIT" },
	{ 0x0007, 0x3123, "LINE TYPE" },
	{ 0x0007, 0x304C, "LOCATE" },
	{ 0x0007, 0x30C9, "LORG" },
    { 0x0007, 0x3372, "MAT AIPLOT" },
    { 0x0007, 0x335C, "MAT APLOT" },
    { 0x0007, 0x337D, "MAT ARPLOT" },
    { 0x0007, 0x3348, "MAT IPLOT" },
    { 0x0007, 0x333E, "MAT PLOT" },
    { 0x0007, 0x3352, "MAT RPLOT" },
    { 0x0007, 0x3367, "MAT SYMBOL" },
    { 0x0007, 0x3321, "MEMORY" },
	{ 0x0007, 0x309C, "MOVE" },
	{ 0x0007, 0x30AD, "MSCALE" },
	{ 0x0007, 0x6660, "OFF GKEY" },
	{ 0x0007, 0x6653, "ON GKEY" },
	{ 0x0007, 0x6646, "ON GKEY CALL" },
	{ 0x0007, 0x662C, "ON GKEY GOTO" },
	{ 0x0007, 0x6639, "ON GKEY GOSUB" },
	{ 0x0007, 0x308C, "PDIR" },
	{ 0x0007, 0x308C, "PDIR" },
	{ 0x0007, 0x3111, "PEN" },
	{ 0x0007, 0x30D9, "PENUP" },
	{ 0x0007, 0x3074, "PLOT" },
	{ 0x0007, 0x302A, "PLOTTER IS" },
	{ 0x0007, 0x3033, "PLOTTER" },
	{ 0x0007, 0x30B6, "POINTER" },
    { 0x0007, 0x332A, "POLYGON" },
    { 0x0007, 0x31FE, "RATIO" },
    { 0x0007, 0x3334, "RECTANGLE" },
	{ 0x0007, 0x3084, "RPLOT" },
	{ 0x0007, 0x303B, "SCALE" },
	{ 0x0007, 0x30FA, "SETGU" },
	{ 0x0007, 0x3102, "SETUU" },
	{ 0x0007, 0x3043, "SHOW" },
	{ 0x0007, 0x6620, "TRACK" },	
	{ 0x0007, 0x30E9, "UNCLIP" },
	{ 0x0007, 0x313E, "WHERE" },
	
	// --- I/O PPU ---
	{ 0x0009, 0x59E1, "COLLECT" },
	{ 0x0009, 0x5936, "OFF INT #1" },
	{ 0x0009, 0x5809, "ON INT #" },
	{ 0x0009, 0x59D2, "SYSTEM TIMEOUT OFF" },
	{ 0x0009, 0x59C2, "SYSTEM TIMEOUT ON" },
	{ 0x0009, 0x406A, "TCLOSE" },
	{ 0x0009, 0x408E, "TDISP" },
	{ 0x0009, 0x4039, "TOPEN" },

	// --- I/O LPU ---
	{ 0x000B, 0x5F4E, "ABORTIO" },
	{ 0x000B, 0x65AC, "CARD ENABLE" },
	{ 0x000B, 0x5E14, "CLEAR" },
	{ 0x000B, 0x5DAC, "CONFIGURE TALK" },
	{ 0x000B, 0x6623, "CONTROL MASK" },
	{ 0x000B, 0x5715, "CONVERT" },
	{ 0x000B, 0x5715, "CONVERT TO" },
	{ 0x000B, 0x601F, "ENTER" },
	{ 0x000B, 0x601F, "ENTER" },
	{ 0x000B, 0x5F7C, "EOI" },
	{ 0x000B, 0x56A3, "EOL" },
	{ 0x000B, 0x5F37, "LOCAL" },
	{ 0x000B, 0x5F25, "LOCAL LOCKOUT" },
	{ 0x000B, 0x549F, "OUTPUT" },
	{ 0x000B, 0x5F9A, "PASS CONTROL" },
	{ 0x000B, 0x5E81, "PPOLL CONFIGURE" },
	{ 0x000B, 0x5E60, "PPOLL UNCONFIGURE" },
	{ 0x000B, 0x6452, "READ IO" },
	{ 0x000B, 0x5E1D, "REMOTE" },
	{ 0x000B, 0x5F68, "REQUEST" },
	{ 0x000B, 0x5DFE, "RESET" },
	{ 0x000B, 0x6646, "SELECT CODE ACTIVE" },
	{ 0x000B, 0x6646, "SELECT CODE INACTIVE" },
	{ 0x000B, 0x5AD3, "SENDBUS" },
	{ 0x000B, 0x6014, "SENTER" },
	{ 0x000B, 0x65C9, "SET TIMEOUT" },
	{ 0x000B, 0x5683, "SOUTPUT" },
	{ 0x000B, 0x653C, "STATUS" },
	{ 0x000B, 0x5E39, "TRIGGER" },
	{ 0x000B, 0x6319, "WAIT READ" },
	{ 0x000B, 0x6445, "WAIT WRITE" },
	{ 0x000B, 0x6488, "WRITE BIN" },
	{ 0x000B, 0x6465, "WRITE IO" },
	{ 0, 0, 0 }
};


void print_usage(char *command)
{
	printf("Usage: %s [<options>] <progfile> [<outfile>]\n", command);
	printf("       <progfile>        PROG input file\n");
	printf("       <outfile>         (optional) output file\n");
	printf("\nwith <options> combination of\n");
	printf("       -binary           input file is a binary program of type BPRG\n");
	printf("       -program          input file is a binary program of type PROG (default)\n");
	printf("       -rom              input file is a ROM image\n");
	printf("       -offset <offset>  skip specified number of words at file start\n");
	printf("       -unsecure         unsecure secured program lines\n");
	printf("       -extract          extract binary and program modules\n");
	printf("       -v                print detailed analyses\n");
	return;
}


void print_log(int level, const char* text, ...)
{
    char tmp[1024];
    va_list args;
    
	if (level>loglevel) return;

    va_start(args, text);
    wvsprintf(tmp, text, args);
    va_end(args);

	printf("%s", tmp);	
}


/*
	Produce a hexdump listing
*/
void dump(int size)
{
  char string[17];
  int i,j,index;
  WORD dummy;
  
  if (size < 1) return;

  LOG(2, "Dumping %d intermediate word(s)...\n", size);
  
  for (i=0; i<size; i++) {
  	READ_WRITE_WORD(dummy);
  	if ((i % 8) == 0) {
  		index = 0;
  		LOG(2, "$%04X: ", offset - 1);
  	}
  	LOG(2, " %04X", dummy);
  	if ((buf[0]>31) && (buf[0]<127)) string[index++] = buf[0];
  	else string[index++] = '.';
  	if ((buf[1]>31) && (buf[1]<127)) string[index++] = buf[1];
  	else string[index++] = '.';
  	if ((i % 8) == 7) {
  		string[index] = 0;
  		LOG(2, " %s\n", string);
  	}
  }
  if ((i % 8) != 0) {
  	for (j=0; j<(8 - (i % 8)); j++)
  		LOG(2, "     ");
	string[index] = 0;
	LOG(2, " %s\n", string);
  }
}

void resolve_opcode(WORD base, WORD opcode)
{
	int target;
	
  	if (opcode == 0) LOG(2, " (NOP)");
  	else if ((opcode & 0xf080) == 0xf080) LOG(2, " (RET %d)", opcode & 0xf);
  	else {
  		target = opcode & 0x3ff;
  		if (target > 511) target -= 1024;
  		switch (opcode & 0xec00) {
		/* current page */
  		case 0xec00:
	  		LOG(2, " (JMP $%04X,I in module)", base + target);
	  		break;
  		case 0x6c00:
	  		LOG(2, " (JMP $%04X in module)", base + target);
	  		break;
		/* base page */
  		case 0xe800:
  			if (target<0) target = 65536 + target;
	  		LOG(2, " (JMP $%04X,I in basepage)", target);
	  		break;
  		case 0x6800:
  			if (target<0) target = 65536 + target;
	  		LOG(2, " (JMP $%04X in basepage)", target);
	  		break;
	  	default:
	  		LOG(2, "(unsupported opcode)");
	  	}
  	}
}

/*
	Process a program module (main program or subprogram)
*/
void parse_program(int size)
{
  int i, module_base, line_offset;
  BYTE column, length;
  WORD dummy, line_number, label_offset, block_address, exec_address;
  BOOL secured, has_label;

  module_base = offset;
  
  LOG(2, "-- Program module --------------------------------------------------------\n");

  while (offset < (module_base + size)) {
  	
  	line_offset = offset;
  	
  	/* line number */
  	READ_WORD(line_number);
  	if (line_number & 0x8000) {
  		if (unsecure) line_number &= 0x7fff;
  		secured = TRUE;
  	}
  	else secured = FALSE;
  	WRITE_WORD(line_number);
  	line_number &= 0x7fff;
  	
  	/* line number valid ? */
    if ((line_number == 0) || (line_number == 32767)) {
    	LOG(2, "Warning: early end of program detected (invalid line number)\n");
    	return;
    };
  	
  	/* column position & line length */
  	READ_WRITE_WORD(dummy);
  	column = buf[0];
  	length = buf[1];
  	
  	READ_WRITE_WORD(dummy);

  	/* has label ? */
  	if (buf[0] & 0x80) {
  		has_label = TRUE;
  		label_offset = (buf[0] * 256 + buf[1]) & 0x7fff;
	  	READ_WRITE_WORD(dummy);
  	}
  	else has_label = FALSE;
  	
  	/* execution address */
  	exec_address = buf[0] * 256 + buf[1];
  	
	LOG(2, "$%04X: %5d col=%2d len=%2d ", line_offset, line_number, column, length);
	LOG(2, "exec=$%04X ", exec_address);
	
	if (exec_address == 0x319) {
		/* option ROM or binary command */

		READ_WRITE_WORD(block_address);
		READ_WRITE_WORD(exec_address);
		
		for (i=0;i<1000;i++) {
			if ((option_table[i].block == 0) && (option_table[i].address == 0)) {
				LOG(2, "<unknown> block=$%04X address=$%04X", block_address, exec_address);
				break;
			}
			if ((option_table[i].block == block_address) && (option_table[i].address == exec_address)) {
				LOG(2, "%s", option_table[i].cmd);
				break;
			}
		}
	}
	else {
		/* standard command */
		for (i=0;i<1000;i++) {
			if (cmd_table[i].address == 0) {
				LOG(2, "<unknown> ");
				break;
			}
			if (cmd_table[i].address == exec_address) {
				LOG(2, "%s", cmd_table[i].cmd);
				break;
			}
		}
	}
	
	if (secured) LOG(2, " (secured)");
//	else LOG(2, "(unsecured)");
	if (has_label) LOG(2, " label=$%04X", module_base + size + label_offset);
	LOG(2, "\n");
	
	/* skip rest of line */
	while (offset < line_offset + length) {
	  	READ_WRITE_WORD(dummy);
	}
  }

  /* skip rest of module */
  while (offset < module_base + size) {
  	READ_WRITE_WORD(dummy);
  }
  
  return;
}

/*
	Parse symbol table
*/
void parse_table(int size)
{
  int i, index, table_base, entry_base, exponent, digits;
  WORD pointer, dummy;
  BYTE type, dim, format;
  BYTE length;
  char identifier[18];
  char mantissa[18];

  LOG(2, "-- Symbol table ----------------------------------------------------------\n");

  table_base = offset;

  while (offset < (table_base + size)) {
  	
  	entry_base = offset;
  	
  	/* get pointer */
  	READ_WRITE_WORD(pointer);
  	
  	/* get type, dim & length*/
  	READ_WRITE_WORD(dummy);
  	type = buf[0] >> 4;
  	dim = buf[0] >> 1 & 0x7;
  	length = buf[1] & 0xf;
  	format = (buf[0] & 0x1) * 16 + ((buf[1] >> 4) & 0xf);

	/* identifier */
	if ((type != 0) && ((type != 1) || (dim != 0))) {
	  	
	  	/* identifier */
		index = 0;
		for (i=offset-entry_base; i<length; i++) {
		  	READ_WRITE_WORD(dummy);
		  	identifier[index++] = buf[0];
		  	identifier[index++] = buf[1];
		}
		identifier[index] = 0;
	}
	
	LOG(2, "$%04X:  offset=$%04X", entry_base, entry_base - table_base);
	if (type != 0) {
		if (pointer != 0x7fff) LOG(2, " pointer=$%04X", pointer);
	}

	switch (type) {
	case 0:
		LOG(2, " REAL constant");
		if (format &0x1) LOG(2, " (float) ");
		else LOG(2, " (fixed) ");
		index = 0;
		digits = format >> 1;
	  	READ_WRITE_WORD(dummy);
		exponent = (buf[0] & 0x7f) * 4 + ((buf[1] & 0xc) >> 6);
		if (buf[0] & 0x80) exponent = exponent - 512;
		if (buf[1] & 0x1) LOG(2, "-");
		for (i=0; i<3; i++) {
		  	READ_WRITE_WORD(dummy);
		  	if (format &0x1) {
		  		/* float */
				mantissa[index++] = ((buf[0]>>4)&0xf) + 0x30;
				if (i == 0) mantissa[index++] = '.';
				mantissa[index++] = (buf[0]&0xf) + 0x30;
				mantissa[index++] = ((buf[1]>>4)&0xf) + 0x30;
				mantissa[index++] = (buf[1]&0xf) + 0x30;
			}
			else {
		  		/* fixed */
				if (index < digits) mantissa[index++] = ((buf[0]>>4)&0xf) + 0x30;
				if (index < digits) mantissa[index++] = (buf[0]&0xf) + 0x30;
				if (index < digits) mantissa[index++] = ((buf[1]>>4)&0xf) + 0x30;
				if (index < digits) mantissa[index++] = (buf[1]&0xf) + 0x30;
			}
		}
		if (((format & 0x1) == 0) && (digits == 0)) mantissa[index++] = '0';
		mantissa[index] = 0;
		LOG(2, "%s", mantissa);
		if (exponent && (format &0x1)) LOG(2, "E%d", exponent);
		//LOG(2, ", %d significant digits", digits);
		break;
	case 1:
		switch (dim) {
		case 0:
		  	READ_WRITE_WORD(dummy);
			LOG(2, " line number %d", dummy);
			break;
		case 1:
			LOG(2, " line label");
			break;
		case 4:
			LOG(2, " numeric function");
			break;
		case 5:
			LOG(2, " STRING function");
			break;
		case 7:
			LOG(2, " subroutine");
			break;
		default:
			LOG(2, " <unknown>");
			break;
		}
		break;
	case 4:
		LOG(2, " STRING variable");
		break;
	case 5:
		LOG(2, " REAL variable");
		break;
	case 6:
		LOG(2, " SHORT variable");
		break;
	case 7:
		LOG(2, " INTEGER variable");
		break;
	default:
		LOG(2, " <unknown>");
		break;
	}
	
	if ((type != 0) && ((type != 1) || (dim != 0))) {
		if ((type != 1) && (dim != 0)) LOG(2, " array (%d dimension(s))", dim);
		LOG(2, " \"%s\"", identifier);
	}
	
	LOG(2, "\n");
  }	
}

/*
	Parse binary program block
*/
void parse_binaries(int size)
{
	int i, module_base, binary_base, index, keyword_base;
	WORD id_info, binary_size, binary_start, flags, num_aux;
	WORD primary_keywords, secondary_keywords, next_keyword, syntax, list, execution, aux, dummy, tmp;
	char id_string[34];
	char keyword[256];
	char filename[MAX_PATH];
	WORD jmptable[4];
	BOOL primary = TRUE;
	
	module_base = offset;
	
	if (is_rom) binary_start = initial_offset;
	
	while (offset < (module_base + size)) {

		LOG(1, "-- Binary program --------------------------------------------------------\n");

		binary_base = offset;
		
		if (extract) {
			/* determine id info offset */
			READ_WORD(id_info);
		  	READ_WORD(binary_size);
			for (i=0; i<6; i++) READ_WORD(dummy);
			READ_WORD(binary_start);

			/* read id info */
			fseek(ifp, (binary_base + id_info - binary_start + 8) * 2, SEEK_SET);
			index = 0;
			for (i=0; i<16; i++) {
			  	READ_WRITE_WORD(dummy);
			  	if (buf[0] & 0x80) break;
			  	if ((buf[0] <= 0x20) || (buf[0] > 0x7e) || strchr(forbidden_characters, buf[0]))
			  		buf[0] ='_';
			  	id_string[index++] = buf[0];
			  	if (buf[1] == 0x80) break;
			  	if ((buf[1] <= 0x20) || (buf[1] > 0x7e) || strchr(forbidden_characters, buf[1]))
			  		buf[1] ='_';
			  	id_string[index++] = buf[1];
			}
			if (id_string[index-1] == ' ') index--;
			id_string[index] = 0;
			
			/* open binary file */
			sprintf(filename, "%s.BPRG", id_string);
			ofp = fopen( filename, "wb");
			if (!ofp) { fprintf(stderr, "Could not open file \"%s\".\n", filename); exit(-1); }
			
			WRITE_WORD(binary_size);
			
			/* reposition file pointer */
			fseek(ifp, binary_base * 2, SEEK_SET);
			offset = binary_base;
		}

		if (is_rom) {
			primary_keywords = binary_base;
			secondary_keywords = 0xffff;
		}
		else {

			/* get binary program header */
		  	READ_WRITE_WORD(id_info);
		  	READ_WRITE_WORD(binary_size);
		  	READ_WRITE_WORD(primary_keywords);
		  	READ_WRITE_WORD(secondary_keywords);
		  	
			/* read jumptable */
			for (i=0; i<4; i++) READ_WRITE_WORD(jmptable[i]);
	
		  	READ_WRITE_WORD(binary_start);	  	
		  	binary_start -= 8;
		  	
		  	/* TODO: check MSB on primary & secondary keyword pointers */
		  	if (binary_start & 0x8000) {
		  		primary_keywords |= 0x8000;
		  		secondary_keywords |= 0x8000;
		  	}
		  	
		  	LOG(2, "$%04X:  ID info offset: $%04X (=$%04X)\n", binary_base, id_info, id_info + binary_base - binary_start);
		  	LOG(2, "$%04X:  Binary size: $%04X (%d) words\n", binary_base + 1, binary_size, binary_size);
		  	LOG(2, "$%04X:  Primary keywords: $%04X (=$%04X)\n", binary_base + 2, primary_keywords, primary_keywords + binary_base - binary_start);
		  	LOG(2, "$%04X:  Secondary keywords: $%04X (=$%04X)\n", binary_base + 3, secondary_keywords, secondary_keywords + binary_base - binary_start);
		  	LOG(2, "$%04X:  On load execute: $%04X", binary_base + 4, jmptable[0]);
		    
		    resolve_opcode(binary_base + 4, jmptable[0]);
		    LOG(2, "\n");

		    if (secondary_keywords <= primary_keywords)
		    	secondary_keywords = binary_start + binary_size;

			for (i=1; i<4; i++) {
				dummy = jmptable[i];
			  	LOG(2, "$%04X:  data=$%04X", binary_base + 4 + i, dummy);
				resolve_opcode(binary_base + 4 + i, dummy);
			  	LOG(2, "\n");
			}
		  	LOG(2, "$%04X:  Binary sync: $%04X (binary start = $%04X)\n", binary_base+8, binary_start+8, binary_start);
	
		  	READ_WRITE_WORD(flags);
		  	LOG(2, "$%04X:  Flags: $%04X\n", offset - 1, flags);
	
		  	READ_WRITE_WORD(num_aux);
		  	LOG(2, "$%04X:  Aux vectors: %d\n", offset - 1, num_aux);
	
			for (i=0; i<num_aux; i++) {
			  	READ_WRITE_WORD(aux);
			  	LOG(2, "$%04X:   - Aux%d: $%04X (=$%04X)\n", offset - 1, i+1, aux, aux - binary_start);
			}
			
			while ((offset - binary_base) < (id_info - binary_start)) {
			  	READ_WRITE_WORD(dummy);
			  	LOG(2, "$%04X:  data=$%04X\n", offset - 1, dummy);
			}
			
			tmp = offset;
			index = 0;
			for (i=0; i<16; i++) {
			  	READ_WRITE_WORD(dummy);
			  	if (buf[0] & 0x80) break;
			  	id_string[index++] = buf[0];
			  	if (buf[1] == 0x80) break;
			  	id_string[index++] = buf[1];
			}
			if (id_string[index-1] == ' ') index--;
			id_string[index] = 0;
			LOG(1, "$%04X:  Binary module ID info: \"%s\"\n", tmp, id_string);
			
			/* skip data */
			dump(primary_keywords - binary_start - (offset - binary_base));
		}
		
		next_keyword = 0;

		/* parse keywords */
		
		LOG(1, "-- Primary keywords --\n");

		while (offset < (binary_base + size)) {
			
			keyword_base = offset;
			
		  	READ_WRITE_WORD(next_keyword);
		  	if (next_keyword == 0xffff) {
		  		LOG(2, "$%04X:  data=$FFFF (terminator)\n", offset - 1);

				/* read the 0x0000 part */
			  	READ_WRITE_WORD(next_keyword);
			  	LOG(2, "$%04X:  data=$%04X\n", offset - 1, next_keyword);

		  		if (!primary || ((offset - binary_base) > (secondary_keywords - binary_start))) break;

		  		primary = !primary;
		  		dump (secondary_keywords - binary_start - (offset - binary_base));
		  		LOG(1, "-- Secondary keywords --\n");
			  	READ_WRITE_WORD(next_keyword);
		  	}

			/* TODO: check MSB next keyword pointer */
		  	if (binary_start & 0x8000) next_keyword |= 0x8000;

	  		LOG(2, "$%04X:  Next keyword: $%04X (=$%04X)\n", offset - 1, next_keyword, next_keyword + binary_base - binary_start);
						
			index = 0;
			tmp = offset;
			while (offset < size) {
			  	READ_WRITE_WORD(dummy);
				if (buf[0] & 0x80) break;
				keyword[index++] = buf[0];
				if (buf[1] == 0x80) { READ_WRITE_WORD(dummy); break; }
				keyword[index++] = buf[1];
			}
			if (keyword[index-1] == ' ') index--;
			keyword[index] = 0;
			
			/* terminator? */
			if (index == 0) {
				LOG(2, "$%04X:  data=$%04X (keyword terminator)\n", offset - 1, dummy);			
				break;
			}
				
			LOG(1, "$%04X:  Keyword=\"%s\"\n", tmp, keyword);
			
			flags = buf[0];
			LOG(2, "$%04X:  - Flags: $%02X (", offset - 1, flags);
			if (flags & 0x40) LOG(2, "prerun, ");
			else LOG(2, "no prerun, ");
			if (flags & 0x20) LOG(2, "no statement, ");
			else LOG(2, "statement, ");
			if (flags & 0x10) LOG(2, "no command, ");
			else LOG(2, "command, ");
			if (flags & 0x08) LOG(2, "cannot be part of IF)\n");
			else LOG(2, "can be part of IF)\n");
			
			if (primary) {			
				execution = tmp + buf[1];
				LOG(2, "$%04X:  - Execution routine: $%02x (=$%04X)\n", offset - 1, buf[1], execution);

			  	READ_WRITE_WORD(syntax);
			  	LOG(2, "$%04X:  - Syntax routine: ", offset - 1);
			  	resolve_opcode(offset - 1, syntax);
			  	LOG(2, "\n");
	
			  	READ_WRITE_WORD(list);
			  	LOG(2, "$%04X:  - List routine: ", offset - 1);
			  	resolve_opcode(offset - 1, list);
			  	LOG(2, "\n");
			}
			
			if (primary && (next_keyword > secondary_keywords)) {
				next_keyword = secondary_keywords;
				primary = FALSE;
			}
				
			/* output data for execution/syntax/list routine*/
			if ((next_keyword + binary_base - binary_start) >= offset) {
				dump(next_keyword - binary_start - (offset - binary_base));
			}
			else {
				dump(size + binary_base - offset);
				break;
			}
			if (next_keyword == secondary_keywords)
				LOG(1, "-- Secondary keywords --\n");

		}
		
		/* output remaining binary module data */
		dump(binary_size + binary_base - offset);
		
		if (extract) {
			WRITE_WORD(binary_size);
			if (ofp) fclose(ofp);
			ofp = 0;
		}

		/* last binary? */
		if ((offset - module_base + 12) > size) {
			LOG(2, "-- End of binary space ---------------------------------------------------\n");
			break;
		}
	}

	/* output remaining data */
	dump(size + module_base - offset);

}

int main(int argc, char *argv[])
{
  int i, count, module_offset, binary_space;
  int address, module_count, module_size, table_offset, table_size, block;
  char *infile = 0;
  char *outfile = 0;
  char filename[MAX_PATH];
  WORD module_table[1024];
  BOOL is_program = FALSE;
  BOOL binary_only = FALSE;
  WORD dummy;

  if (argc < 2) { print_usage(argv[0]); exit(0); }

  /* command line processing */
  for ( count = 1; count < argc ; count++)
  {
  	
	if (!strcmp(argv[count], "-offset"))
		{ count++; initial_offset = strtoul(argv[count],0,0); continue; }
  	
	if (!strcmp(argv[count], "-binary"))
		{ binary_only = TRUE; continue; }

	if (!strcmp(argv[count], "-program"))
		{ is_program = TRUE; continue; }

	if (!strcmp(argv[count], "-rom"))
		{ is_rom = TRUE; continue; }

	if (!strcmp(argv[count], "-unsecure"))
		{ unsecure = TRUE; continue; }

	if (!strcmp(argv[count], "-extract"))
		{ extract = TRUE; continue; }

	if (!strcmp(argv[count], "-v"))
		{ loglevel = 2; continue; }

	if (!strncmp(argv[count], "-", 1) && (count < argc))
		 { print_usage(argv[0]); exit(0); }

	if (!infile) { infile = argv[count]; continue; }
	if (!outfile) { outfile = argv[count]; continue; }
  }
  
  if (!infile) { print_usage(argv[0]); exit(0); }
  	
  /* normal or binary program? */
  printf("[%s]\n", infile + strlen(infile) - 5);
  if (!strcmp(infile + strlen(infile) - 5, ".BPRG")) {
	if (is_program) LOG(1, "Warning: file %s appears to be a binary program\n", infile);
	else binary_only = TRUE;
  } 
  
  if (!strcmp(infile + strlen(infile) - 5, ".PROG")) {
	if (binary_only) LOG(1, "Warning: file %s appears to be a normal program\n", infile);
	else is_program = TRUE;
  }
  
  if (is_program && binary_only) binary_only = FALSE;

  /* open PROG input & output files */
  ifp = fopen( infile, "rb");
  if (!ifp) { fprintf(stderr, "Could not open file \"%s\".\n", infile); exit(-1); }
  if (!extract && outfile) {
	  ofp = fopen( outfile, "wb");
	  if (!ofp) { fprintf(stderr, "Could not open file \"%s\".\n", outfile); exit(-1); }
  }

  LOG(1, "-- File header -----------------------------------------------------------\n");

  fseek(ifp, 0, SEEK_SET);
  
  if (initial_offset) {
	for (i=0; i<initial_offset; i++)
		READ_WRITE_WORD(dummy);
  }
  
  if (is_rom) binary_space = 0xffff;
  else {
	  READ_WRITE_WORD(binary_space);
	  if (binary_space)
	  	LOG(1, "$%04X:  Binary space: $%04X (%d) words\n", offset - 1, binary_space, binary_space);  	
	  else
	  	LOG(1, "$%04X:  File does not contain binary programs\n", offset - 1);
	  
	  if (!binary_only) {
		  READ_WRITE_WORD(module_count);
		  
		  LOG(1, "$%04X:  Program has %d program module(s):\n", offset - 1, module_count);
		  
		  if (module_count > 1024) {
		  	LOG(1, "Warning: max. 1024 modules supported\n");
		  	module_count = 1024;
		  }
		  
		  /* read in module table */
		  for (i=0; i<module_count; i++) {
		    READ_WRITE_WORD(module_table[i]);
		  	LOG(1, "$%04X:  - Module #%2i: $%04X (%d) words\n", offset - 1, i+1, module_table[i], module_table[i]);
		  }
	  }
  }

  if (binary_space) {
  	parse_binaries(binary_space);	
	if (!extract) {
		READ_WRITE_WORD(dummy);
		LOG(2, "$%04X:  - Binary space: $%04X (%d) words\n", offset - 1, dummy, dummy);
		if (dummy != binary_space) LOG(2, "Warning: binary space size mismatch\n");
	}
  }

  if (binary_only || is_rom || (module_count == 0)) {
	  if (ifp) fclose(ifp);
	  if (ofp) fclose(ofp);
	  exit(0);
  }
  
  if (extract) {
		/* open PROG file */
		if (outfile) strcpy(filename, outfile);
		else sprintf(filename, "%s.out", infile);
		ofp = fopen(filename, "wb");
		if (!ofp) { fprintf(stderr, "Could not open file \"%s\".\n", filename); exit(-1); }
			
		/* write file header */
		WRITE_WORD(0);
		WRITE_WORD(module_count);
		for (i=0; i<module_count; i++)
			WRITE_WORD(module_table[i]);
  }
  	  	
  for (i=0; i<module_count; i++) {
  	
  	LOG(2, "Parsing module #%d\n", i);
  	
  	/* read module header */

	LOG(2, "-- Module header ---------------------------------------------------------\n");

    READ_WRITE_WORD(module_size);
  	LOG(2, "$%04X:  Module size: $%04X (%d) words\n", offset - 1, module_size, module_size);
  	if (module_size != module_table[i])
  		LOG(2, "Warning: module size (%d) doesn't match table entry (%d)\n", module_size, module_table[i]);

    READ_WRITE_WORD(table_offset);
  	LOG(2, "$%04X:  Symbol table offset: $%04X (%d) words\n", offset - 1, table_offset, table_offset);

    READ_WRITE_WORD(table_size);
  	LOG(2, "$%04X:  Symbol table size: %d words\n", offset - 1, table_size);

    READ_WRITE_WORD(address);
  	LOG(2, "$%04X:  Address in memory block: $%04X\n", offset - 1, address);
  	
    READ_WRITE_WORD(block);
  	LOG(2, "$%04X:  Memory block number: $%04X\n", offset - 1, block);

  	parse_program(table_offset);

  	parse_table(table_size);
  	
  	dump(module_size - table_size - table_offset - 5);  	
  }
  
  /* final check */
  READ_WRITE_WORD(i);
  LOG(2, "$%04X:  Terminating module count: %d\n", offset - 1, i);
  if (i != module_count)
  	LOG(2, "Warning: terminating module count (=%d) doesn't match initial count (=%d)\n", i, module_count);

  if (extract) {
  	if (ofp) fclose(ofp);
  	ofp = 0;
  }
  
  LOG(2, "-- End of program module space -------------------------------------------\n");

  /* ouput remaining stuff */
  module_size = 0;
  module_offset = offset;
  while(fread(buf, 1, 2, ifp)) module_size++;
  fseek(ifp, module_offset, SEEK_SET);
  dump(module_size);
  
  /* cleanup */
  if (ifp) fclose(ifp);
  if (ofp) fclose(ofp);
  exit(0);
}
